# -*- coding: utf-8 -*-
"""TBA."""

import warnings
import numpy as np
import synphot

from astropy.io import fits
from astropy import units as u
from astropy.wcs import WCS
from astropy.utils.decorators import deprecated_renamed_argument

from synphot import SourceSpectrum, Empirical1D

from spextra import Spextrum
from ..rc import Source, ter_curve_utils as tu
from ..utils.general_utils import add_function_call_str, make_img_wcs_header,\
    RA0, DEC0


__all__ = ["point_source",
           "uniform_source",
           "source_from_array",
           "source_from_file",
           "source_from_imagehdu",
           "source_from_imagehdu_with_flux",
           "source_from_cube",
           ]


def _make_spectrum(sed):
    if isinstance(sed, (Spextrum, SourceSpectrum)):
        return Spextrum(modelclass=sed)
    try:
        return Spextrum(sed)
    except TypeError as err:
        raise TypeError("Parameter 'sed' must be Spextrum, SourceSpectrum or "
                        f"str, found {type(sed)=}.") from err


def _make_scaled_spectrum(sed, amplitude, filter_curve):
    spec = _make_spectrum(sed)
    if amplitude is None:
        return spec

    # This might be redundant, check spextra method!
    if not isinstance(amplitude, u.Quantity):
        amplitude = amplitude * u.ABmag

    return spec.scale_to_magnitude(amplitude, filter_curve)


@add_function_call_str
def point_source(sed, amplitude=None, filter_curve="V", x=0, y=0,
                 ra=RA0, dec=DEC0) -> Source:
    """
    Create a point source with an arbitrary spectrum.

    This is similar to ``scopesim_templates.stellar.star``.

    Parameters
    ----------
    sed : str or synphot.Source_Spectrum
        Passing a str will try to download a sed from the speXtra database
        alternatively an user manipulated ``synphot.Source_Spectrum`` or
        compatible object can be provided.
    amplitude : float
        flux or magnitude of the object. The SED will be scaled to that value
        if left to `None` (default) no scaling will be performed
    filter_curve : str
        any astronomical filter in the speXtra or the spanish VO database
    x : float
        X-coordinate on the plane in arcsec
    y : float
        Y-coordinate on the plane in arcsec
    ra : float, str
        RA coordinates of the center of the field (not used)
    dec : float, str
        DEC coordinates of the center of the field  (not used)

    Return
    ------
    src : ``scopesim.Source``
    """
    spec = _make_scaled_spectrum(sed, amplitude, filter_curve)
    src = Source(spectra=spec, x=[x], y=[y], ref=[0], weight=[1])
    return src


@add_function_call_str
def uniform_source(sed, amplitude=None, filter_curve="V", extend=60,
                   ra=RA0, dec=DEC0) -> Source:
    """
    Create an extended uniform source with an arbitrary spectrum.

    This function reates an image with extend^2 pixels with size of 1 arcsec^2
    so provided amplitudes.
    are in flux or magnitudes per arcsec^2

    Parameters
    ----------
    sed : synphot or spextra sed
    amplitude : float
        magnitude or flux (PER ARCSEC^2) of the spectrum in the specified
        filter_curve.
    filter_curve : str
        any filter curve
    extend : int
        extension of the field in arcsec, will always produce a square field
    ra : float
        RA of the field center (not used)
    dec : float
        DEC of field center (not used)

    Return
    ------

    src : ``scopesim.Source``
    """
    data = np.ones(shape=(extend, extend))
    header = make_img_wcs_header(ra=ra, dec=dec, pixel_scale=1,
                                 image_size=data.shape)
    hdu = fits.ImageHDU(header=header, data=data)

    spec = _make_scaled_spectrum(sed, amplitude, filter_curve)
    src = Source(spectra=spec, image_hdu=hdu)
    return src


@add_function_call_str
def source_from_imagehdu(image_hdu, filter_name, pixel_unit_amplitude=None,
                         waverange=None, inst_pkg_path=None) -> Source:
    r"""
    Create a ``scopesim.Source`` object directly from a ``fits.ImageHDU``.

    Parameters
    ----------
    image_hdu : fits.ImageHDU
    filter_name : str
        Either a standard filter name or a filter from an instrument package
    waverange: tuple
        wave_min and wave_max of the spectral range
    pixel_unit_amplitude, optional
        A Quantity that corresponds to a pixel value of 1 in the image
        If not given, header keyword BUNIT is used
    inst_pkg_path : str, optional
        Not yet implemented

    Returns
    -------
    src : scopesim.Source

    Examples
    --------
    Using a generic filter curve from the Spanish VO::

        >>> image_hdu = fits.ImageHDU(data=np.ones((11, 11)))
        >>> # add WCS info to the header here
        >>> filter_name = "Generic/Johnson_UBVRIJHKL.N"
        >>> src = misc.source_from_imagehdu(image_hdu=hdu,
                                            filter_name=filter_name,
                                            pixel_unit_amplitude=20*u.Jy)

    Using the METIS H2O-ice filter from the METIS ScopeSim package::

        >>> import scopesim
        >>> filter_name = scopesim.rc.__search_path__[0] + \
                          "/METIS/filters/TC_filter_H2O-ice.dat"
        >>> src = misc.source_from_imagehdu(image_hdu=hdu,
                                            filter_name=filter_name,
                                            pixel_unit_amplitude=20*u.Jy)

    TODO: Check if the image_hdu has WCS
    """
    if filter_name is None and waverange is None:
        raise ValueError("Wavelength information must be given with either a "
                         "filter_name or a waverange")

    units = image_hdu.header.get("BUNIT")
    amp = 1.

    if isinstance(pixel_unit_amplitude, u.Quantity):
        amp = pixel_unit_amplitude.value
        units = pixel_unit_amplitude.unit
    elif isinstance(pixel_unit_amplitude, (list, np.ndarray)):
        amp = pixel_unit_amplitude

    if units is None:
        raise ValueError("Units must be supplied with either the BUNIT keyword"
                         "or as an astropy.Quantity with pixel_unit_amplitude")

    amp_unit = amp * u.Unit(units)

    if filter_name is not None and isinstance(filter_name, str):
        filter_curve = tu.get_filter(filter_name)
        waverange = filter_curve.waverange

    waves = np.linspace(waverange[0], waverange[1], num=10)
    lookup_table = [1,] * len(waves)
    spec = SourceSpectrum(Empirical1D, points=waves, lookup_table=lookup_table)
    spec = tu.scale_spectrum(spec, filter_name=filter_name, amplitude=amp_unit)

    if image_hdu.header.get("SPEC_REF") is None:
        image_hdu.header["SPEC_REF"] = 0

    src = Source(image_hdu=image_hdu, spectra=spec)
    return src


@add_function_call_str
def source_from_imagehdu_with_flux(image_hdu=None, filename=None, ext=1,
                                   pixel_scale=None, flux=None,
                                   bunit=None) -> Source:
    """
    Create a source from an image where pixel values have known flux and unit.

    The units of the flux density expressed by 'bunit'.

    It is possible to change the flux and pixel scale to simulate e.g. more
    distant objects.

    Parameters
    ----------
    image_hdu : fits.ImageHDU
        ImageHDU instance or
    filename : str
        A fits filename.
    ext : int
        Extension where the data and header is located in the file.
    pixel_scale : float, optional
        The pixel scale in arcsec if not in the headers or a different one is
        needed.
    flux : float, u.Quantity, optional
        The total flux of the image if the user wants to specify a different
        one than in the image.
    bunit : u.Quantity
        The units of flux if BUNIT is not in the headers or a different one is
        needed.

    Returns
    -------
    src : scopesim.Source


    Note
    ----
    Creates a source with a flat spectrum. It is responsibility of the user to
    use it correctly with the proper filter and instrument configuration.

    """
    if image_hdu is not None:
        header = image_hdu.header
        data = image_hdu.data
    elif filename is not None:
        header = fits.getheader(filename, ext=ext)
        data = fits.getdata(filename, ext=ext)
    else:
        raise ValueError("Please provide a filename or an ImageHDU")

    # we need this because ScopeSim doesn't understand sky coordinates
    header.update({"CRVAL1": 0, "CRVAL2": 0})

    if pixel_scale is not None:
        pixel_scale /= 3600
        header.update({"CDELT1": pixel_scale, "CDELT2": pixel_scale})

    if bunit is not None:
        unit = u.Unit(bunit).to_string()
        header.update({"BUNIT": unit})

    if isinstance(flux, u.Quantity):
        unit = flux.unit.to_string()
        header.update({"BUNIT": unit})
        flux = flux.value
    elif flux is None:
        flux = np.sum(data)

    unit = u.Unit(header["BUNIT"])
    total_flux = flux * unit
    data = data / np.sum(data)
    image_hdu = fits.ImageHDU(data=data, header=header)
    src = Source(image_hdu=image_hdu, flux=total_flux)
    return src


@add_function_call_str
def source_from_array(arr, sed, pixel_scale, amplitude, filter_curve,
                      ra=RA0, dec=DEC0) -> Source:
    """
    Create a source from an image (numpy 2D array).

    Parameters
    ----------
    arr : np.ndarray
      a 2D numpy array
    sed : basestring
       any sed available in the spextra library or a Spextrum object
    pixel_scale : float, u.Quantity
        pixel scale on the sky (u.arcsec, u.arcmin, u.deg).
    amplitude : float, u.Quantity
       magnitude of the spectra in a filter_curve
    filter_curve : basestring
        a filter in the speXtra database or the SVO

    Returns
    -------
    src : ``scopesim.Source``
    """
    pixel_scale <<= u.arcsec

    header = make_img_wcs_header(ra=ra, dec=dec, pixel_scale=pixel_scale,
                                 image_size=arr.shape)
    header.update({"SPEC_REF": 0})

    data = arr / np.sum(arr)
    hdu = fits.ImageHDU(data=data, header=header)

    spec = _make_scaled_spectrum(sed, amplitude, filter_curve)
    src = Source(image_hdu=hdu, spectra=spec)
    return src


@deprecated_renamed_argument(old_name="image", new_name="arr", since="0.3.0",
                             arg_in_kwargs=True)
def source_from_image(**kwargs):
    warnings.warn("source_from_image has been replaced by source_from_array."
                  "This is to avoid confusion with source_from_imagehdu",
                  PendingDeprecationWarning)
    return source_from_array(**kwargs)


source_from_image.__doc__ = source_from_array.__doc__


@add_function_call_str
def source_from_file(filename, pixel_scale, sed, amplitude, filter_curve,
                     cut=0, ext=1, **kwargs) -> Source:
    """
    Create a source from a fits image.

    TODO: Pass the header (or the WCS) to the source object

    Parameters
    ----------
    filename
    pixel_scale
    sed
    amplitude
    filter_curve
    ext : int
       fits extension
    cut
    **kwargs : arguments passed to fits.open
    Returns
    -------
    """
    hdu = fits.open(filename, **kwargs)
    data = hdu[ext].data
#    header = hdu[ext].header
#    wcs = WCS(header)
#    if wcs.has_celestial:
#        ra, dec = wcs.wcs.crval

    data[data < cut] = 0
    src = source_from_array(arr=data, sed=sed, pixel_scale=pixel_scale,
                            amplitude=amplitude, filter_curve=filter_curve)

    return src


def source_from_cube(cube, ext=1) -> Source:
    """
    Wrapper for Source(cube=cube) for completeness.

    In the future more checks will be performed to translate most cubes into a
    suitable format.

    TODO: Set the cube to a different flux level and different pixel scales.
    Redshift the cube

    Parameters
    ----------
    cube : Filename, ImageHDU or HDUList containing the cube
    ext : extension where the data is located

    """
    src = Source(cube=cube, ext=ext)

    return src


def poorman_cube_source(filename=None, hdu=None, ext=1, pixel_scale=None,
                        amplitude=None, filter_curve=None) -> Source:
    """
    Create a very basic source from cube.

    That source might work with the current implementation of ``field_of_view``
    objects.

    It basically divides the cube in monocromatic slices (images).
    The associated spectra are just delta functions.
    """
    if filename is not None:
        header = fits.getheader(filename, ext=ext)
        data = fits.getdata(filename, ext=ext)
    elif hdu is not None:
        header = hdu[ext].header
        data = hdu[ext].data
    else:
        raise ValueError("Please define a datacube")

    try:
        bunit = u.Unit(header["BUNIT"])
    except KeyError as err:
        raise ValueError("No BUNIT defined") from err

    if pixel_scale is not None:  # assuming lineal
        try:
            pixel_scale = pixel_scale.to(u.deg).value
        except AttributeError:
            pixel_scale = pixel_scale / 3600

    wcs = WCS(header)

    celwcs = wcs.celestial
#    celwcs.wcs.cd = np.array([[-1*pixel_scale, 0], [0, pixel_scale]])  # needs more thought

    celhdr = celwcs.to_header()

    specwcs = wcs.spectral

    zpix = np.arange(specwcs.spectral.array_shape[0])
    # keeping wavelengths in native coordinates
    waves = specwcs.pixel_to_world(zpix)
    assert len(waves) > 3, "At least 4 wavelength points are required."
    zero_spec = np.zeros(shape=zpix.shape) * bunit

    hdus = []
    specs = []

    src = Source()
    for i in zpix:
        imgdata = data[i]
        flux = np.sum(imgdata)
        imgdata = imgdata / flux

        celhdr.update({"SPEC_REF": i})
        hdus.append(fits.ImageHDU(data=imgdata, header=celhdr))

        zero_spec[i] = flux * bunit
        spec = synphot.SourceSpectrum(synphot.Empirical1D, points=waves,
                                      lookup_table=zero_spec)
        specs.append(spec)
        zero_spec[i] = 0

        src = src + Source(image_hdu=fits.ImageHDU(data=imgdata,
                                                   header=celhdr),
                           spectra=spec)


#    src = Source(image_hdu=hdul, spectra=specs)

    return src


def source_from_moments():
    """Place holder for that function."""
    raise NotImplementedError()



import pytest

import pyckles

from scopesim_templates.stellar import stars_utils as stu


@pytest.fixture(scope="class")
def pickles():
    return pyckles.SpectralLibrary("pickles")


@pytest.mark.usefixtures("pickles")
class TestNearestSpecType:
    @pytest.mark.parametrize("spec_types", ("A0V", "G2V",
                                            "O8III", "G5III_R"))
    def test_returns_existing_spec_type(self, pickles, spec_types):
        spec = stu.nearest_spec_type(spec_types, pickles.table)
        assert spec == spec_types

    @pytest.mark.parametrize("input, output", (["K7I", "K7V"],
                                               ["G2II", "G2I"],
                                               ["G2III", "G2IV"]))
    def test_returns_nearest_evolutionary_type(self, pickles, input, output):
        spec = stu.nearest_spec_type(input, pickles.table)
        assert spec == output

    @pytest.mark.parametrize("input, output", (["O1I", "O5V"],
                                               ["G1II", "G0I"],
                                               ["F4V", "F5V"]))
    def test_returns_nearest_luminosity_class(self, pickles, input, output):
        spec = stu.nearest_spec_type(input, pickles.table)
        assert spec == output

    def test_returns_list_for_list_input(self, pickles):
        list_in = ["O1I", "G1II", "F4V"]
        list_out = ["O5V", "G0I", "F5V"]
        spec = stu.nearest_spec_type(list_in, pickles.table)
        assert all([so == lo for so, lo in zip(spec, list_out)])

    def test_throws_error_if_wierd_stuff_inputted(self, pickles):
        with pytest.raises(ValueError):
            stu.nearest_spec_type("bogus", pickles.table)



import pytest

import numpy as np

from astropy.io import fits
import astropy.units as u


def _basic_imagehdu(n=11):
    hdr_dict = {"CRPIX1": (n + 1) / 2,
                "CRPIX2": (n + 1) / 2,
                "CRVAL1": 0,
                "CRVAL2": 0,
                "CDELT1": 0.1 / 3600,
                "CDELT2": 0.1 / 3600,
                "CUNIT1": "deg",
                "CUNIT2": "deg",
                "CTYPE1": 'RA---TAN',
                "CTYPE2": 'DEC--TAN'}

    im = np.zeros((n,n))
    for x,y in zip([0, 0, -1, -1, 5], [0, -1, 0, -1, 5]):
        im[0] = 1

    hdu = fits.ImageHDU(data=im)
    hdu.header.update(hdr_dict)

    return hdu


@pytest.fixture(scope="module")
def starting_star_imagehdu():
    """Create the same star as in starting.ipynb"""
    n = 100
    sigma = 5
    x, y = np.meshgrid(np.arange(n), np.arange(n))
    img = np.exp(-1 * (((x - n / 2) / sigma) ** 2 + ((y - n / 2) / sigma) ** 2))

    # Fits headers of the image. Yes it needs a WCS
    hdr = fits.Header(dict(NAXIS=2,
                           NAXIS1=n,
                           NAXIS2=n,
                           CRPIX1=(n + 1) / 2,
                           CRPIX2=(n + 1) / 2,
                           CRVAL1=0,
                           CRVAL2=0,
                           CDELT1=0.2 / 3600,
                           CDELT2=0.2 / 3600,
                           CUNIT1="DEG",
                           CUNIT2="DEG",
                           CTYPE1='RA---TAN',
                           CTYPE2='DEC--TAN'))

    # Creating an ImageHDU object
    hdu = fits.ImageHDU(data=img, header=hdr)
    return hdu


def _make_dummy_cube(scale, wave_unit, ref_wave, wave_step, wave_type, bunit):

    if isinstance(scale, u.Quantity) is False:
        scale = scale * u.arcsec
    if isinstance(wave_unit, u.core.Unit) is False:
        wave_unit = u.AA
    if isinstance(ref_wave, u.Quantity) is False:
        ref_wave = ref_wave * u.AA
    if isinstance(wave_step, u.Quantity) is False:
        wave_step = wave_step * u.AA

    data = np.ones(shape=(100, 20, 20))
    header = fits.Header(dict(NAXIS=3,
                              WCSAXES=3,
                              NAXIS1=data.shape[2] + 1,
                              NAXIS2=data.shape[1] + 1,
                              NAXIS3=data.shape[0] + 1,
                              CRPIX1=(data.shape[2] + 1) / 2,
                              CRPIX2=(data.shape[1] + 1) / 2,
                              CRPIX3=1,
                              CRVAL1=0,
                              CRVAL2=0,
                              CRVAL3=ref_wave.to(wave_unit).value,
                              CDELT1=-1 * scale.to(u.deg).value,
                              CDELT2=scale.to(u.deg).value,
                              CDELT3=wave_step.to(wave_unit).value,
                              CUNIT1="deg",
                              CUNIT2="deg",
                              CUNIT3=wave_unit.to_string(),
                              CTYPE1='RA---TAN',
                              CTYPE2='DEC--TAN',
                              CTYPE3=wave_type,
                              BUNIT=bunit))

    hdu = fits.ImageHDU(data=data, header=header)

    return hdu



# -*- coding: utf-8 -*-
"""WCU laser spectrum for METIS."""

import numpy as np
from astropy import units as u
from astropy.io import fits

from spextra import Spextrum

from ..utils.general_utils import add_function_call_str
from ..rc import Source, im_plane_utils


@add_function_call_str
def laser_spectrum(
        centers,
        fwhms,
        fluxes,
        img_size=(12, 12),
        waves=None,
        specdict=None,
):
    """
    Create laser calibration spectrum source for METIS.

    Parameters
    ----------
    centers : TYPE
        Emission line centers.
    fwhms : TYPE
        Emission line FWHMs.
    fluxes : TYPE
        Emission line total fluxes.
    img_size : TYPE, optional
        FOV size in arcsec, plus a little extra. The default is (12, 12).
    waves : array-like, optional
        Input wavelength axis. Either `waves` or `specdict` must be given. If
        both are given, `waves` takes precedence.The default is None.
    specdict : Mapping, optional
        Dict-like structure containing the keys "wave_min", "wave_max",
        "spectral_bin_width" and "wave_unit", which is used to construct the
        wavelength axis. Either `waves` or `specdict` must be given. If both
        are given, `waves` takes precedence. The default is None.

    Returns
    -------
    src : Source

    """
    if waves is None:
        if specdict is None:
            raise ValueError("Either waves or specdict must be given.")
        waves = np.arange(
            specdict["wave_min"],
            specdict["wave_max"],
            specdict["spectral_bin_width"]
        ) * u.Unit(specdict["wave_unit"])

    # Taken from micado.spectral_calibration.line_list:
    dw = 0.5 * img_size[0] / 3600         # input needed in [deg]
    dh = 0.5 * img_size[1] / 3600         # input needed in [deg]
    pixel_scale = 0.1 * dw

    hdr = im_plane_utils.header_from_list_of_xy(
        [-dw, dw], [-dh, dh], pixel_scale)  # [deg]
    hdr["SPEC_REF"] = 0
    im = np.ones((hdr["NAXIS1"], hdr["NAXIS2"]))
    field = fits.ImageHDU(header=hdr, data=im)
    spec = Spextrum.emission_line_spectrum(centers.round(5), fwhms, fluxes,
                                           waves=waves)
    src = Source(image_hdu=field, spectra=spec)
    return src


def laser_spectrum_lm(specdict, n_tunable: int = 20, **kwargs):
    """
    Calibration laser for METIS LM bands.

    Create a fixed red and blue laser as well as a variable number of tunable
    laser lines in between.

    Parameters
    ----------
    specdict : dict-like
        Dict-like structure containing the keys "wave_min", "wave_max",
        "spectral_bin_width" and "wave_unit", which is used to construct the
        wavelength axis. Usually this means ``cmd["!SIM.spectral"]``.
    n_tunable : int, optional
        Number of emission lines created for the tunable laser between 4.68 and
        4.78 um. Passing 1 will create a single line at 4.73 um.
        The default is 20.
    **kwargs : TYPE
        Any other keyword arguments passed to ``laser_spectrum``.
        In particular, `centers`, `fwhms` and `fluxes` will override the
        default values and any lines created by `n_tunable`. If those are
        overridding, make sure the shapes of those three arguments match.

    Returns
    -------
    src : Source

    Notes
    -----
    The default values for the emission lines are 3.39 and 5.263 um for the
    fixed lasers, with FWHM of 8E-8 and 6E-5 um respectively. The various lines
    created by the tunable laser all use the same FWHM of 8E-7 um. The fluxes
    of all lines are scaled to create lines of roughly equal height.

    """
    if n_tunable == 1:
        tunable_cen = [4.73] * u.um
    else:
        tunable_cen = np.linspace(4.68, 4.78, n_tunable) * u.um

    defaults = {
        "centers": [3.39, *tunable_cen.value, 5.263] * u.um,
        "fwhms": [8E-8, *[8E-7]*len(tunable_cen), 6E-5] * u.um,
        "fluxes": [1, *[7]*len(tunable_cen), 500],
        "specdict": specdict
    } | kwargs
    return laser_spectrum(**defaults)


def laser_spectrum_n(specdict, **kwargs):
    """
    Calibration laser for METIS N band.

    Create equal emission lines at 9, 10, 11 and 12 um.

    Parameters
    ----------
    specdict : dict-like
        Dict-like structure containing the keys "wave_min", "wave_max",
        "spectral_bin_width" and "wave_unit", which is used to construct the
        wavelength axis. Usually this means ``cmd["!SIM.spectral"]``.
    **kwargs : TYPE
        Any other keyword arguments passed to ``laser_spectrum``.

    Returns
    -------
    src : Source

    Notes
    -----
    The default of 1E-8 um is used as a best-guess for the line's FWHMs.
    The relative fluxes are scaled to produce lines of roughly equal height.
    """
    defaults = {
        "centers": [9, 10, 11, 12] * u.um,
        "fwhms": 4 * [1E-8] * u.um,
        "fluxes": [1.3, 1.2, 1.1, 1],
        "specdict": specdict
    } | kwargs
    return laser_spectrum(**defaults)
